#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Simple Post-Execution Message with information about RAM and Time."""


import atexit
import logging as log
import os
import sys

from datetime import datetime

try:
    import resource
except ImportError:
    resource = None  # MS Window dont have resource


def make_post_exec_msg(start_time=None, comment=None):
    """Simple Post-Execution Message with information about RAM and Time."""
    use, al, msg = 0, 0, ""
    if sys.platform.startswith("windows"):
        msg = "No information about RAM usage available on non-Linux systems\n"
    elif sys.platform.startswith("linux"):
        use = int(resource.getrusage(resource.RUSAGE_SELF).ru_maxrss *
                  resource.getpagesize() / 1024 / 1024 if resource else 0)
        al = int(os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES') /
                 1024 / 1024 if hasattr(os, "sysconf") else 0)
        msg += "Total Maximum RAM Memory used: {0} of {1}MegaBytes.\n".format(
            use, al)
        if start_time and datetime:
            msg += "Total Working Time: ~ {0}.\n".format(
                datetime.now() - start_time)
    if comment:
        msg += str(comment).strip()
    log.debug("Preparing Simple Post-Execution Messages.")
    atexit.register(log.info, msg)
    return msg



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Install handler attach post-mortem pdb console on an exception."""


import logging as log
import sys
import traceback


def pdb_on_exception(debugger="pdb", limit=100):
    """Install handler attach post-mortem pdb console on an exception."""
    log.debug("Installing an automatic Debugger upon Exceptions...")

    def pdb_excepthook(exc_type, exc_val, exc_tb):
        traceback.print_tb(exc_tb, limit=limit)
        __import__(str(debugger).strip().lower()).post_mortem(exc_tb)

    sys.excepthook = pdb_excepthook


def ipdb_on_exception(debugger="ipdb", limit=100):
    """pdb_on_exception but using iPDB instead."""
    return pdb_on_exception("ipdb", limit=limit)


# ptpdb_on_exception = lambda: pdb_on_exception("ptpdb")
# See https://github.com/jonathanslenders/ptpdb/issues/14



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Import a module from file path string."""


import errno
import importlib.util
import logging as log
import os

from .exceptions import NamespaceConflictError


def path2import(pat, name=None, ignore_exceptions=False, check_namespace=True):
    """Import a module from file path string.

    This is "as best as it can be" way to load a module from a file path string
    that I can find from the official Python Docs, for Python 3.5+.
    """
    module = None
    if not os.path.isfile(pat):
        if not ignore_exceptions:
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), pat)
    elif not os.access(pat, os.R_OK):
        if not ignore_exceptions:
            raise PermissionError(pat)
    else:
        try:
            name = name or os.path.splitext(os.path.basename(pat))[0]
            if check_namespace and name in globals():
                e = "Module {0} already exist on global namespace".format(name)
                if not ignore_exceptions:
                    raise NamespaceConflictError(e)
            else:
                spec = importlib.util.spec_from_file_location(name, pat)
                if spec is None:
                    e = 'Failed to load module {0} from {1}.'.format(name, pat)
                    if not ignore_exceptions:
                        raise ImportError(e)
                module = spec.loader.load_module()
        except Exception as error:
            log.warning("Failed to Load Module {0} from {1}".format(name, pat))
            log.warning(error)
            if not ignore_exceptions:
                raise
            module = None
        else:
            log.debug("Loading Module {0} from path {1}.".format(name, pat))
    return module



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Get current public IP as string."""


from urllib.request import urlopen


def get_public_ip():
    """Get current public IP as string."""
    return urlopen("https://api.ipify.org").read().decode("utf-8").strip()


def is_online():
    """Check if we got internet conectivity."""
    try:
        get_public_ip()
    except:
        return False
    else:
        return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Add to autostart or launcher icon on the Desktop."""


import logging as log
import os
import sys


def set_desktop_launcher(app, desktop_file_content, autostart=False):
    """Add to autostart or launcher icon on the Desktop."""
    if not sys.platform.startswith("linux"):
        return  # .desktop files are Linux only AFAIK.
    # Auto-Start file below.
    config_dir = os.path.join(os.path.expanduser("~"), ".config", "autostart")
    os.makedirs(config_dir, exist_ok=True)
    autostart_file = os.path.join(config_dir, app + ".desktop")
    if os.path.isdir(config_dir) and not os.path.isfile(autostart_file):
        if autostart:
            log.info("Writing Auto-Start Executable file: " + autostart_file)
            with open(autostart_file, "w", encoding="utf-8") as start_file:
                start_file.write(desktop_file_content)
        if os.path.isfile(autostart_file):
            os.chmod(autostart_file, 0o776)
    # Desktop Launcher file below.
    apps_dir = os.path.join(os.path.expanduser("~"),  # paths are XDG.
                            ".local", "share", "applications")
    os.makedirs(apps_dir, exist_ok=True)
    desktop_file = os.path.join(apps_dir, app + ".desktop")
    if os.path.isdir(apps_dir) and not os.path.isfile(desktop_file):
        log.info("Writing Desktop Launcher Executable file: " + desktop_file)
        with open(desktop_file, "w", encoding="utf-8") as desktop_file_obj:
            desktop_file_obj.write(desktop_file_content)
    if os.path.isfile(desktop_file):
        os.chmod(desktop_file, 0o776)
    return desktop_file



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Format number as currency money with thousand separator, 2 decimals."""


def number2currency(value, decimals=2, sign="$"):
    """Format number as currency money with thousand separator, 2 decimals."""
    number, decimal = (('%%,%df' % decimals) % value).split(',')
    parts = []
    while len(number) > 3:
        part, number = number[-3:], number[:-3]
        parts.append(part)
    parts.append(number)
    parts.reverse()
    if int(decimal) == 0:
        currency_money = '.'.join(parts) + sign
    else:
        currency_money = '.'.join(parts) + ',' + decimal + sign
    return currency_money



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Make a Chainable concurrent.futures.Future that has a .then() api."""


import sys
import weakref

from concurrent.futures import Future


class ChainableFuture(Future):

    """Make a Chainable concurrent.futures.Future that has a .then() api."""

    @property
    def _chained_future_log(self):
        """Chained Future Log."""
        prop_name = '_chained_futured_log_list'
        log = getattr(self, prop_name, None)
        if log is None:
            log = weakref.WeakSet()
            setattr(self, prop_name, log)
        return log

    def _chain_to_another_future(self, base_future):
        """Chain a Future instance directly to another Future instance."""
        msg = "Circular chain error.Future {} is already in resolved chain {}."
        if base_future in self._chained_future_log:
            raise Exception(msg.format(base_future,
                                       set(self._chained_future_log)))
        else:
            self._chained_future_log.add(base_future)

        def _done_handler(base_future):
            """Convert return of previous Future to results of new Future."""
            if not base_future.done():  # Never True. Avoid infinite timeout.
                self.cancel()
                return
            if base_future.cancelled():
                self.cancel()
                return
            try:
                result = base_future.result()
                if isinstance(result, Future):
                    self._chain_to_another_future(result)
                else:
                    self.set_result(result)
                return
            except BaseException:
                ex, trace_back = sys.exc_info()[1:]
                self.set_exception_info(ex, trace_back)
                return
        base_future.add_done_callback(_done_handler)

    def then(self, on_success=None, on_fail=None):
        """Chainable concurrent.futures.Future.then() JS-like.

        - on_success: Optional,Callable to run when this Future success Ok.
        - on_fail: Optional,Callable to run when this Future fails.
        - Return: Future chained to current Future.
        """
        new_future = self.__class__()

        def _done_handler(base_future):
            """Convert result of previous Future to result of new Future."""
            if not base_future.done():  # Never True. Avoid infinite timeout.
                new_future.cancel()
                return
            if base_future.cancelled():
                new_future.cancel()
                return
            try:
                result = base_future.result()
                if on_success:
                    result = on_success(result)
                if isinstance(result, Future):  # Defer resolution new_future
                    new_future._chain_to_another_future(result)
                else:
                    new_future.set_result(result)
                    return
            except BaseException:
                ex, trace_back = sys.exc_info()[1:]
                if not on_fail:
                    new_future.set_exception_info(ex, trace_back)
                    return
                else:
                    try:
                        result = on_fail(ex)
                        if isinstance(result, BaseException):
                            raise result
                        else:
                            new_future.set_result(result)
                        return
                    except BaseException:
                        ex, trace_back = sys.exc_info()[1:]
                        new_future.set_exception_info(ex, trace_back)
                        return
        self.add_done_callback(_done_handler)
        return new_future



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Debug and Log Encodings and Check for root/administrator,return Bool."""


import logging as log
import os
import sys

from getpass import getuser
from platform import platform, python_version


def check_encoding(check_root=True):
    """Debug and Log Encodings and Check for root/administrator,return Bool."""
    log.debug("Python {0} on {1}.".format(python_version(), platform()))
    log.debug("STDIN Encoding: {0}.".format(sys.stdin.encoding))
    log.debug("STDERR Encoding: {0}.".format(sys.stderr.encoding))
    log.debug("STDOUT Encoding:{}".format(getattr(sys.stdout, "encoding", "")))
    log.debug("Default Encoding: {0}.".format(sys.getdefaultencoding()))
    log.debug("FileSystem Encoding: {0}.".format(sys.getfilesystemencoding()))
    log.debug("PYTHONIOENCODING Encoding: {0}.".format(
        os.environ.get("PYTHONIOENCODING", None)))
    os.environ["PYTHONIOENCODING"] = "utf-8"
    if sys.platform.startswith(("linux", "darwin")) and check_root:  # root
        if not os.geteuid():
            log.warning("Runing as root is not Recommended !.")
            return False
    elif sys.platform.startswith("windows") and check_root:  # administrator
        if getuser().lower().startswith("admin"):
            log.warning("Runing as Administrator is not Recommended !.")
            return False
    return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Set process I/O and cpu priority."""


import logging as log
import os

from shutil import which
from subprocess import call


def set_process_priority(nice=True, ionice=False):
    """Set process name and cpu priority."""
    w = "IONice may delay I/O Operations, not recommended on user-facing GUI!."
    try:
        if nice:
            old = os.getpriority(os.PRIO_PROCESS, 0)
            os.nice(19)  # smooth cpu priority
            log.debug("Process CPUs Priority set: from {0} to 19.".format(old))
        elif ionice and which("ionice"):
            log.warning(w)
            command = "{0} --ignore --class 3 --pid {1}".format(
                which("ionice"), os.getpid())
            call(command, shell=True)  # I/O nice,should work on Linux and Os X
            log.debug("Process I/O Priority set to: {0}.".format(command))
    except Exception as error:
        log.warning(error)
        return False  # this may fail on windows and its normal, so be silent.
    else:
        return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Tiny Template Engine that Render and Runs native Python."""
# Renamed from Templar to TemplatePython for easy of use.
# Was about to use TemplateString, but will confuse with string.Template


import re


class TemplatePython(str):

    """Tiny Template Engine that Render and Runs native Python."""

    def __init__(self, template):
        """Init the Template class."""
        self.tokens = self.compile(template.strip())

    @classmethod
    def from_file(cls, fl):
        """Load template from file.A str/file-like object supporting read()."""
        return cls(str(open(fl).read() if isinstance(fl, str) else fl.read()))

    def compile(self, t):
        """Parse and Compile all Tokens found on the template string t."""
        tokens = []
        for i, p in enumerate(re.compile("\{\%(.*?)\%\}", re.DOTALL).split(t)):
            if not p or not p.strip():
                continue
            elif i % 2 == 0:
                tokens.append((False, p.replace("{\\%", "{%")))
            else:
                lines = tuple(p.replace("%\\}", "%}").replace(
                    "{{", "spit(").replace("}}", "); ") .splitlines())
                mar = min(len(_) - len(_.lstrip()) for _ in lines if _.strip())
                al = "\n".join(line_of_code[mar:] for line_of_code in lines)
                tokens.append((True, compile(al, "<t {}>".format(al), "exec")))
        return tokens

    def render(__self, __namespace={}, mini=False, **kw):
        """Render template from __namespace dict + **kw added to namespace."""
        html = []
        __namespace.update(kw, **globals())

        def spit(*args, **kwargs):
            for _ in args:
                html.append(str(_))
            if kwargs:
                for _ in tuple(kwargs.items()):
                    html.append(str(_))

        __namespace["spit"] = spit
        for is_code, value in __self.tokens:
            eval(value, __namespace) if is_code else html.append(value)
        return re.sub('>\s+<', '> <', "".join(html)) if mini else "".join(html)

    __call__ = render  # shorthand



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Get and return a free unused port."""


import logging as log
import socket


def get_free_port(port_range=(8000, 9000)):
    """Get and return a free unused port."""
    sockety = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sockety.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    for port_number in range(int(port_range[0]), int(port_range[1])):
        try:
            sockety.bind(("127.0.0.1", port_number))
        except (socket.error, Exception):
            pass
        else:
            sockety.close()
            del sockety
            log.debug("Found free unused port number: {0}".format(port_number))
            return port_number



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Stealth Strings, hidden and dangerous."""


import base64
import binascii
import codecs
import zlib


def string2stealth(stringy: str, rot13: bool=False) -> str:
    """String to Stealth,stealth is a hidden string,both str type and utf-8."""
    stringy = codecs.encode(stringy, "rot-13") if rot13 and codecs else stringy
    strng = base64.b64encode(zlib.compress(stringy.strip().encode('utf-8'), 9))
    bits = bin(int(binascii.hexlify(strng), 16))[2:]
    return str(bits.zfill(8 * ((len(bits) + 7) // 8))).replace(
        "0", u"\u200B").replace("1", u"\uFEFF")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Make a JSON from Nested to Flat with an arbitrary delimiter."""


def make_json_flat(jsony, delimiter="__"):
    """Make a JSON from Nested to Flat with an arbitrary delimiter."""
    values = {}
    for item in jsony.keys():
        if isinstance(jsony[item], dict):
            get = make_json_flat(jsony[item], delimiter)
            for something in get.keys():
                values[item + delimiter + something] = get[something]
        else:
            values[item] = jsony[item]
    return values



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Add ENV environtment variables to python globals dict."""


import logging as log
import os


def env2globals(pattern="PY_"):
    """Add ENV environtment variables to python globals dict."""
    log.debug("Adding environtment variables that match {0} to globals".format(
              pattern))
    try:
        for var in [_ for _ in os.environ.items() if _[0].startswith(pattern)]:
            globals().update({var[0]: var[1]})  # tuple to dict
    except Exception as e:
        log.warning(e)
        return False
    else:
        return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Set process name and cpu priority,return socket.socket or None."""


import logging as log
import socket
import sys


def set_single_instance(name, port=8888):
    """Set process name and cpu priority,return socket.socket or None."""
    try:  # Single instance app ~crossplatform, uses udp socket.
        log.info("Creating Abstract UDP Socket Lock for Single Instance.")
        __lock = socket.socket(
            socket.AF_UNIX if sys.platform.startswith("linux")
            else socket.AF_INET, socket.SOCK_STREAM)
        __lock.bind(
            "\0_{name}__lock".format(name=str(name).lower().strip())
            if sys.platform.startswith("linux") else ("127.0.0.1", port))
    except (socket.error, OSError, Exception) as e:
        __lock = None
        log.critical("Another instance of App is already running!, Exiting!.")
        log.exception(e)
        sys.exit()
        exit()
    else:
        log.info("Socket Lock for Single Instance: {}.".format(__lock))
    finally:
        return __lock



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Common exceptions for anglerfish."""


class AnglerfishException(Exception):
    """Common exceptions for anglerfish."""
    pass


class NamespaceConflictError(ImportError, AnglerfishException):
    """Common exceptions for anglerfish."""
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Convert bytes to kilobytes, megabytes, gigabytes, etc."""


import logging as log


def bytes2human(bites, to, bsize=1024):
    """Convert bytes to kilobytes, megabytes, gigabytes, etc."""
    log.debug("Converting {0} Bytes to {1}.".format(bites, to.upper()))
    bitez, to = float(abs(bites)), str(to).strip().lower()
    for i in range({'k': 1, 'm': 2, 'g': 3, 't': 4, 'p': 5, 'e': 6}[to]):
        bitez = bitez / int(abs(bsize))
    return str(int(bitez)) + {
        'k': " Kilo", 'm': " Mega", 'g': " Giga", 't': " Tera",
        'p': " Peta", 'e': " Exa"}[to] + "bytes"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Check if running on Battery, return Bool."""


import logging as log
import os

try:
    import dbus
except ImportError:
    dbus = None


def _get_prop(obj, iface, prop):
    """Get object interface properties."""
    if not dbus:
        log.warning("D-Bus module not found or not supported on this OS.")
        return  # Windows probably.
    try:
        return obj.Get(iface, prop, dbus_interface=dbus.PROPERTIES_IFACE)
    except (dbus.DBusException, dbus.exceptions.DBusException) as err:
        print(err)
        return


def has_battery():
    """Check if we are connected to a AC power or Battery."""
    log.debug("Checking if connected to AC-Power or Battery.")
    battery_path = "/sys/class/power_supply"  # is it universal on Linux ?
    if not os.path.exists(battery_path):
        return False
    for folder in os.listdir(battery_path):
        type_path = os.path.join(battery_path, folder, 'type')
        if os.path.exists(type_path):
            with open(type_path) as power_file:
                if power_file.read().lower().startswith('battery'):
                    return True
    return False


def on_battery():
    """Check if we are running on Battery power."""
    log.debug("Checking if running on Battery power.")
    if has_battery():
        bus, upower_path = dbus.SystemBus(), '/org/freedesktop/UPower'
        upower = bus.get_object('org.freedesktop.UPower', upower_path)
        result = _get_prop(upower, upower_path, 'OnBattery')
        if result is None:  # Cannot read property, something is wrong.
            print("Failed to read D-Bus property: {0}.".format(upower_path))
            result = False  # Assume we are connected to a power supply.
        return result
    return False

#
# def on_wifi():
#     """Check if we are running on wifi."""
#     try:
#         bus = dbus.SystemBus()
#         manager = bus.get_object('org.freedesktop.NetworkManager',
#                                  '/org/freedesktop/NetworkManager')
#          # FIXME this returns int, I dunno what they mean ?, investigate.
#         return _get_prop(manager, 'org.freedesktop.NetworkManager',
#                          'WirelessEnabled')
#     except Exception:
#         return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Anglerfish."""


import atexit
import logging
import os
import sys
import signal
import zipfile

from logging.handlers import TimedRotatingFileHandler
from copy import copy
from datetime import datetime
from tempfile import gettempdir
from random import choice

try:
    import resource
except ImportError:
    resource = None  # MS Window dont have resource


##############################################################################


from anglerfish.check_encoding import check_encoding  # noqa
from anglerfish.check_folder import check_folder  # noqa
from anglerfish.get_clipboard import get_clipboard  # noqa
from anglerfish.get_sanitized_string import get_sanitized_string  # noqa
from anglerfish.make_beep import beep  # noqa
from anglerfish.make_json_pretty import json_pretty  # noqa
from anglerfish.make_log_exception import log_exception  # noqa
from anglerfish.make_multiprocess import multiprocessed  # noqa
from anglerfish.make_multithread import threads  # noqa
from anglerfish.make_postexec_message import make_post_exec_msg  # noqa
from anglerfish.make_retry import retry  # noqa
from anglerfish.make_typecheck import typecheck  # noqa
from anglerfish.walk2list import walk2list  # noqa
from anglerfish.make_watch import watch  # noqa
from anglerfish.set_desktop_launcher import set_desktop_launcher  # noqa
from anglerfish.set_process_name import set_process_name  # noqa
from anglerfish.set_single_instance import set_single_instance  # noqa
from anglerfish.get_temp_folder import get_temp_folder  # noqa
from anglerfish.set_terminal_title import set_terminal_title  # noqa
from anglerfish.bytes2human import bytes2human  # noqa
from anglerfish.walk2dict import walk2dict  # noqa
from anglerfish.seconds2human import seconds2human, timedelta2human  # noqa
from anglerfish.env2globals import env2globals  # noqa
from anglerfish.html2ebook import html2ebook  # noqa
from anglerfish.make_template_python import TemplatePython  # noqa
from anglerfish.get_free_port import get_free_port  # noqa
from anglerfish.path2import import path2import  # noqa
from anglerfish.make_notification import make_notification  # noqa
from anglerfish.make_json_flat import make_json_flat  # noqa
from anglerfish.json2xml import json2xml  # noqa
from anglerfish.check_hardware import has_battery, on_battery  # noqa
from anglerfish.make_zip_comment import set_zip_comment, get_zip_comment  # noqa
from anglerfish.set_display_off import set_display_off  # noqa
from anglerfish.make_chainable_future import ChainableFuture  # noqa

from anglerfish.get_pdb_on_exception import (pdb_on_exception,  # noqa
                                             ipdb_on_exception)  # noqa
from anglerfish.make_test_terminal_color import make_test_terminal_color
from anglerfish.string2stealth import string2stealth  # noqa
from anglerfish.stealth2string import stealth2string  # noqa
from anglerfish.get_public_ip import get_public_ip, is_online  # noqa
from anglerfish.set_process_priority import set_process_priority  # noqa
from anglerfish.number2currency import number2currency  # noqa
from anglerfish.get_random_pastel_color import (get_random_pastelight_color,
                                                get_random_pasteldark_color,
                                                get_random_pastel_color)


##############################################################################


__version__ = '2.0.0'
__license__ = ' GPLv3+ LGPLv3+ '
__author__ = ' Juan Carlos '
__email__ = ' juancarlospaco@gmail.com '
__url__ = 'https://github.com/juancarlospaco/anglerfish'
__all__ = (
    'ChainableFuture', 'TemplatePython',  # Those are Classes.
    'beep', 'bytes2human', 'check_encoding', 'check_folder', 'env2globals',
    'get_clipboard', 'get_free_port', 'get_sanitized_string',  # All functions.
    'get_temp_folder', 'get_zip_comment', 'has_battery', 'html2ebook',
    'ipdb_on_exception', 'json2xml', 'json_pretty', 'log_exception',
    'make_json_flat', 'make_logger', 'make_notification', 'make_post_exec_msg',
    'multiprocessed', 'on_battery', 'path2import', 'pdb_on_exception',
    'retry', 'seconds2human', 'set_desktop_launcher',
    'set_display_off', 'set_process_name', 'set_single_instance',
    'set_terminal_title', 'set_zip_comment', 'start_time', 'threads',
    'timedelta2human', 'typecheck', 'walk2dict', 'walk2list', 'watch',
    'make_test_terminal_color', 'string2stealth', 'stealth2string',
    'get_public_ip', 'is_online', 'set_process_priority', 'number2currency',
    'get_random_pastelight_color', 'get_random_pasteldark_color',
    'get_random_pastel_color',
    'AnglerfishException', 'NamespaceConflictError'  # Exceptions
)

sys.dont_write_bytecode, start_time = True, datetime.now()
signal.signal(signal.SIGINT, signal.SIG_DFL)


##############################################################################


def __zip_old_logs(log_file, single_zip):
    zip_file, filename = log_file + "s-old.zip", os.path.basename(log_file)
    log.debug("ZIP Compressing Unused Old Rotated Logs.")
    comment = "Compressed Unused Old Rotated Python Logs since ~{0}.".format(
        datetime.now().isoformat()[:-7])
    logs = [os.path.join(os.path.dirname(log_file), _)
            for _ in os.listdir(os.path.dirname(log_file))
            if ".log." in _ and not _.endswith(".zip") and filename in _]
    if single_zip:  # If 1 ZIP for all Logs, put all *.log inside 1 *.zip
        with zipfile.ZipFile(zip_file, 'a', zipfile.ZIP_DEFLATED) as log_zip:
            log_zip.debug = 3  # Log ZIP inner working,and comment with time
            log_zip.comment = bytes(comment, encoding="utf-8")  # add a comment
            for fyle in logs:
                try:
                    log_zip.write(fyle, os.path.basename(fyle))
                    os.remove(fyle)
                except:
                    pass
            log_zip.printdir()
    else:  # If not 1 ZIP, put 1 *.log inside 1 *.zip, multiple zips
        for fyle in logs:
            newzip = fyle + ".zip"
            with zipfile.ZipFile(newzip, 'w', zipfile.ZIP_DEFLATED) as log_zip:
                log_zip.debug = 3  # Log ZIP inner working
                log_zip.comment = bytes(comment, encoding="utf-8")
                try:
                    log_zip.write(fyle, os.path.basename(fyle))
                    os.remove(fyle)
                except:
                    pass
                # log_zip.printdir()
    result = zip_file if single_zip else tuple([_ + ".zip" for _ in logs])
    log.debug(result)
    return result


def make_logger(name, when='midnight', single_zip=False,
                log_file=None, backup_count=100, emoji=False):
    """Build and return a Logging Logger."""
    global log
    if not log_file:
        log_file = os.path.join(gettempdir(), name.lower().strip() + ".log")
    atexit.register(__zip_old_logs, log_file, single_zip)  # ZIP Old Logs
    hand = TimedRotatingFileHandler(log_file, when=when,
                                    backupCount=backup_count, encoding="utf-8")
    hand.setLevel(-1)
    _fmt = ("%(asctime)s %(levelname)s: "
            "%(processName)s (%(process)d) %(threadName)s (%(thread)d) "
            "%(name)s.%(funcName)s: %(message)s %(pathname)s:%(lineno)d")
    hand.setFormatter(logging.Formatter(fmt=_fmt, datefmt="%Y-%m-%d %H:%M:%S"))
    log = logging.getLogger()
    log.addHandler(hand)
    log.setLevel(-1)
    if not sys.platform.startswith("win") and sys.stderr.isatty():
        log.debug("Enabled Colored Logs on current Terminal.")
        make_test_terminal_color()

        def add_color_emit_ansi(fn):
            """Add methods we need to the class."""
            def new(*args):
                """Method overload."""
                if len(args) == 2:
                    new_args = (args[0], copy(args[1]))
                else:
                    new_args = (args[0], copy(args[1]), args[2:])
                if hasattr(args[0], 'baseFilename'):
                    return fn(*args)
                levelno, end = new_args[1].levelno, ' \x1b[0m'
                if levelno >= 50:
                    color = '\x1b[31;5;7m\n '  # blinking red with black
                    if emoji:
                        end += choice((' 😿 \n', ' 🙀 \n', ' 💩 \n', ' ☠ \n'))
                elif levelno >= 40:
                    color = '\x1b[31m'  # red
                    if emoji:
                        end += choice((' 😾 ', ' 😼 ', ''))
                elif levelno >= 30:
                    color = '\x1b[33m'  # yellow
                    if emoji:
                        end += choice((' 😺 ', ' 😻 ', ''))
                elif levelno >= 20:
                    color = '\x1b[32m'  # green
                    if emoji:
                        end += choice((' 😸 ', ' 😽 ', ''))
                elif levelno >= 10:
                    color = '\x1b[35m'  # pink
                    if emoji:
                        end += choice((' 🐱 ', ' 😹 ', ''))
                else:
                    color = '\x1b[0m'  # normal
                try:
                    new_args[1].msg = color + str(new_args[1].msg) + end
                except Exception as reason:
                    print(reason)  # Do not use log here.
                return fn(*new_args)
            return new
        logging.StreamHandler.emit = add_color_emit_ansi(
            logging.StreamHandler.emit)

    log.addHandler(logging.StreamHandler(sys.stderr))
    if os.path.exists("/dev/log") or os.path.exists("/var/run/syslog"):
        is_linux = sys.platform.startswith("linux")
        adrs = "/dev/log" if is_linux else "/var/run/syslog"
        try:
            handler = logging.handlers.SysLogHandler(address=adrs)
            handler.setFormatter(logging.Formatter(
                fmt=_fmt, datefmt="%Y-%m-%d %H:%M:%S"))
        except Exception:
            log.debug("Unix SysLog Server not found,ignore Logging to SysLog")
        else:
            log.addHandler(handler)
            log.debug("Unix SysLog Server trying to Log to SysLog: " + adrs)
    log.debug("Logger created with Log file at: {0}.".format(log_file))
    return log



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Multithreading helper."""


from concurrent.futures import ThreadPoolExecutor
from functools import wraps


class _Threaded():

    """Basic Threaded class."""

    def __init__(self, future, timeout):
        """Init _Threaded class, set class attributes."""
        self._future, self._timeout = future, timeout

    def __getattr__(self, name):
        """Get and return the name attribute."""
        result = self._wait()
        return result.__getattribute__(name)

    def _wait(self):
        """Get wait."""
        return self._future.result(self._timeout)


def _async(n, base_type, timeout=None):
    """Async internal function for decorator."""
    def decorator(f):
        """Decorator builder."""
        if isinstance(n, int):
            pool = base_type(n)
        elif isinstance(n, base_type):
            pool = n
        else:
            raise TypeError("Invalid Type: {}".format(type(base_type)))

        @wraps(f)
        def wrapped(*args, **kwargs):
            """Return the wrapped function."""
            return _Threaded(pool.submit(f, *args, **kwargs), timeout=timeout)
        return wrapped

    return decorator


def threads(n, timeout=None):
    """Convert a simple function to multrithreading."""
    return _async(n, ThreadPoolExecutor, timeout)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Calculate time, with precision from seconds to days."""


def seconds2human(time_on_seconds=0, do_year=True,
                  unit_words={"y": " Years ", "d": " Days ", "h": " Hours ",
                              "m": " Minutes ", "s": " Seconds "}):
    """Calculate time, with precision from seconds to days."""
    minutes, seconds = divmod(int(abs(time_on_seconds)), 60)
    hours, minutes = divmod(minutes, 60)
    days, hours = divmod(hours, 24)
    human_time_string = ""
    yea = "{0}" + str(unit_words.get("y", " Years "))
    dayz = "%02d" + str(unit_words.get("d", " Days "))
    hou = "%02d" + str(unit_words.get("h", " Hours "))
    minu = "%02d" + str(unit_words.get("m", " Minutes "))
    seco = "%02d" + str(unit_words.get("s", " Seconds "))
    years = int(int(days) / 365)
    if days and years and do_year and years > 1:
            human_time_string += yea.format(years)
            days = days - (365 * years)
    if days:
        human_time_string += dayz % days
    if hours:
        human_time_string += hou % hours
    if minutes:
        human_time_string += minu % minutes
    human_time_string += seco % seconds
    return human_time_string.strip()


def timedelta2human(time_delta, do_year=True,  # Just a shortcut :)
                    unit_words={"y": " Years ", "d": " Days ", "h": " Hours ",
                                "m": " Minutes ", "s": " Seconds "}):
    """Convert a TimeDelta object to human string representation."""
    return seconds2human(time_delta.total_seconds(),
                         do_year=do_year, unit_words=unit_words)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Set Monitor Display OFF, return Bool."""


import logging as log
import sys
from shutil import which
from subprocess import call


def set_display_off():
    """Set Monitor Display OFF, it should Auto-ON when needed, return Bool."""
    log.debug("Setting Monitor Display OFF.")
    if sys.platform.startswith('linux') and which("xset"):
        return not bool(call(which("xset") + " dpms force off", shell=True))
    elif sys.platform.startswith('darwin'):
        return not bool(call(
            """echo 'tell application "Finder" to sleep' | osascript""",
            shell=True))
    elif sys.platform.startswith('win'):  # Complicated as hell,dont worth it.
        log.warning("Set Monitor Display OFF not supported on this OS.")
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Pretty-Printing JSON data from dictionary to string."""


import logging as log

from json import dumps


def json_pretty(json_dict: dict) -> str:
    """Pretty-Printing JSON data from dictionary to string."""
    log.debug("Pretty-Printing JSON data string...")
    _json = dumps(json_dict, sort_keys=1, indent=4, separators=(",\n", ": "))
    posible_ends = tuple('true false , " ] 0 1 2 3 4 5 6 7 8 9 \n'.split(" "))
    max_indent, justified_json = 1, ""
    for json_line in _json.splitlines():
        if len(json_line.split(":")) >= 2 and json_line.endswith(posible_ends):
            lenght = len(json_line.split(":")[0].rstrip()) + 1
            max_indent = lenght if lenght > max_indent else max_indent
            max_indent = max_indent if max_indent <= 80 else 80  # Limit indent
    for line_of_json in _json.splitlines():
        is_str = line_of_json.strip().endswith(('"', '",')) and \
            line_of_json.strip().startswith('"') and ":" in line_of_json
        condition_1 = max_indent > 1 and len(line_of_json.split(":")) >= 2
        condition_2 = line_of_json.endswith(posible_ends) and len(line_of_json)
        if condition_1 and condition_2 and not is_str:
            propert_len = len(line_of_json.split(":")[0].rstrip()) + 1
            xtra_spaces = " " * (max_indent + 1 - propert_len)
            xtra_spaces = ":" + xtra_spaces
            justified_line_of_json = ""
            justified_line_of_json = line_of_json.split(":")[0].rstrip()
            justified_line_of_json += xtra_spaces
            justified_line_of_json += "".join(
                line_of_json.split(":")[1:len(line_of_json.split(":"))])
            justified_json += justified_line_of_json + "\n"
        else:
            justified_json += line_of_json + "\n"
    return str("\n\n" + justified_json if max_indent > 1 else _json)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Multiprocessing helper."""


import multiprocessing
import sys

from concurrent import futures


class __MultiProcessed():

    """Basic MultiProcessed class."""

    def __init__(self, cpu_num, thread_num):
        self.cpu_num, self.thread_num = cpu_num, thread_num

    def _multi_cpu(self, _func, job_queue, timeout):
        if _getLen(job_queue) == 0:
            return []
        index = _get_index(job_queue, self.cpu_num)
        cpu_pool = multiprocessing.Pool(processes=self.cpu_num)
        mgr = multiprocessing.Manager()
        process_bar = mgr.list()
        for i in range(self.cpu_num):
            process_bar.append(0)
        result_queue = cpu_pool.map(
            _multi_thread,
            [[_func, self.cpu_num, self.thread_num, job_queue[int(index[i][0]):
              int(index[i][1] + 1)], timeout, process_bar, i]
             for i in range(len(index))])
        result = []
        for rl in result_queue:
            for r in rl:
                result.append(r)
        return result


def _func(argv):
    argv[2][argv[3]] = round((argv[4] * 100.0 / argv[5]), 2)
    sys.stdout.write(str(argv[2]) + ' ||' + '->' + "\r")
    sys.stdout.flush()
    return argv[0](argv[1])


def _multi_thread(argv):
    thread_num = argv[2]
    if _getLen(argv[3]) < thread_num:
        thread_num = argv[3]
    _func_argvs = [[argv[0], argv[3][i], argv[5], argv[6], i, len(argv[3])]
                   for i in range(len(argv[3]))]
    result = []
    if thread_num == 1:
        for _func_argv in _func_argvs:
            result.append(_func(_func_argv))
        return result
    thread_pool = futures.ThreadPoolExecutor(max_workers=thread_num)
    result = thread_pool.map(_func, _func_argvs, timeout=argv[4])
    return [r for r in result]


def _getLen(_list):
    return 0 if _list is None else len(_list)


def _get_index(job_queue, split_num):
    job_num = _getLen(job_queue)
    if job_num < split_num:
        split_num = job_num
    each_num = job_num / split_num
    index = [[i * each_num, i * each_num + each_num - 1]
             for i in range(split_num)]
    residual_num = job_num % split_num
    for i in range(residual_num):
        index[split_num - residual_num + i][0] += i
        index[split_num - residual_num + i][1] += i + 1
    return index


def multiprocessed(function, arguments, cpu_num=1, thread_num=1, timeout=None):
    """Run on multiple processes and threads the given callable."""
    multicpu_instance = __MultiProcessed(cpu_num, thread_num)
    return multicpu_instance._multi_cpu(function, arguments, timeout)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Test the Terminal Colors."""


def make_test_terminal_color():
    """Test the Terminal Colors."""
    print("Testing Terminal Colors...")
    r = g = b = 0
    result = ""
    for r in range(81):
        result += '\x1b[0;48;2;%s;%s;%sm ' % (r, g, b)
    result += "\n"
    r = 0
    for g in range(81):
        result += '\x1b[0;48;2;%s;%s;%sm ' % (r, g, b)
    result += "\n"
    g = 0
    for b in range(81):
        result += '\x1b[0;48;2;%s;%s;%sm ' % (r, g, b)
    result += '\x1b[0m \n'
    print(result)
    return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Log Exceptions but pretty printing with more info, return string."""


import logging as log
import sys
import traceback


def log_exception():
    """Log Exceptions but pretty printing with more info, return string."""
    unfriendly_names = {"<module>": "Unnamed Anonymous Module Function",
                        "<stdin>": "System Standard Input Function"}
    line_tpl = "    |___ {key} = {val}  # Type: {t}, Size: {s}Bytes, ID: {i}\n"
    body_tpl = """
    ################################ D E B U G ###############################
    Listing all Local objects by context frame, ordered by innermost last:
    {body}
    Thats all we know about the error, check the LOG file and StdOut.
    ############################### D E B U G #############################"""
    tb, body_txt, whole_txt = sys.exc_info()[2], "", ""
    while 1:
        if not tb.tb_next:
            break
        tb = tb.tb_next
    stack = []
    f = tb.tb_frame
    while f:
        stack.append(f)
        f = f.f_back
    stack.reverse()
    traceback.print_exc()
    for frame in stack:
        if frame.f_code.co_name in unfriendly_names.keys():
            fun = unfriendly_names[frame.f_code.co_name]
        else:
            fun = "Function {0}()".format(frame.f_code.co_name)
        body_txt += "\nThe {nm} from file {fl} at line {ln} failed!.".format(
            nm=fun, fl=frame.f_code.co_filename, ln=frame.f_lineno)
        body_txt += "\n    {}\n    |\n".format(fun)
        for key, value in frame.f_locals.items():
            whole_txt += line_tpl.format(key=key, val=repr(value)[:50],
                                         t=str(type(value))[:25],
                                         s=sys.getsizeof(key), i=id(value))
    result = body_tpl.format(body=body_txt + whole_txt)
    log.debug(result)
    return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Perform full walk of where, gather full path of all files."""


import logging as log
import os

from typing import NamedTuple


def walk2list(where, target, omit, links=False, tuply=True, namedtuple=None):
    """Perform full walk of where, gather full path of all files."""
    log.debug("Scan {},searching {},ignoring {},{}following SymLinks.".format(
        where, target, omit, "" if links else "Not "))
    listy = [os.path.abspath(os.path.join(r, f))
             for r, d, fs in os.walk(where, followlinks=links)
             for f in fs if not f.startswith('.') and
             not f.endswith(omit) and
             f.endswith(target)]  # only target files,no hidden files
    list_of_files = listy
    if tuply:  # Return a Tuple.
        list_of_files = tuple(listy)
    if namedtuple:  # Return a NamedTuple with static typing.
        namedtuple_of_files = NamedTuple(  # You can work comfortably using
            str(namedtuple).strip(),  # myfolder.path_9 instead of myfolder[9]
            fields=(("path_{0}".format(i), str) for i in range(len(listy))))
        list_of_files = namedtuple_of_files(*listy)
    return list_of_files



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Retry calling the decorated function using an exponential backoff."""


import functools
import time


def retry(tries=5, delay=3, backoff=2, timeout=None,
          silent=False, logger=None, exceptions=(Exception, )):
    """Retry calling the decorated function using an exponential backoff."""
    def deco_retry(f):

        @functools.wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = int(tries), int(delay)
            end_time = time.time() + int(timeout) if timeout else None
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except exceptions as e:
                    if end_time and time.time() > end_time:
                        raise
                    msg = "{0}, Retrying in {1} seconds...".format(e, mdelay)
                    if mtries < 3:
                        msg += " (Warning: This is the last Retry!)."
                    if not silent:
                        logger.warning(msg) if logger else print(msg)
                    time.sleep(mdelay + int(mtries % 2))
                    mtries -= 1  # Backoff multiplier should be between 1 ~ 8
                    mdelay *= int(backoff if backoff in range(1, 9) else 1)
            return f(*args, **kwargs)

        return f_retry

    return deco_retry



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Convert JSON to XML."""


def json2xml(json_obj, line_padding=""):
    """Convert JSON to XML."""
    result_list, json_obj_type = [], type(json_obj)

    if json_obj_type is list:
        for sub_elem in json_obj:
            result_list.append(json2xml(sub_elem, line_padding))
        return "\n".join(result_list)

    if json_obj_type is dict:
        for tag_name in json_obj:
            sub_obj = json_obj[tag_name]
            result_list.append("{0}<{1}>".format(line_padding, tag_name))
            result_list.append(json2xml(sub_obj, "    " + line_padding))
            result_list.append("{0}</{1}>".format(line_padding, tag_name))
        return "\n".join(result_list)

    return "{0}{1}".format(line_padding, json_obj)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Set or Get a ZIP comment on its Metadata, return Bool."""


import logging as log

from zipfile import ZipFile


def set_zip_comment(zip_path, comment=""):
    """Set a ZIP comment."""
    try:
        with ZipFile(str(zip_path).strip(), 'a') as myzip:
            myzip.comment = bytes(str(comment).strip().encode("utf-8"))
    except Exception as error:
        log.warning("Failed to set comment to ZIP file: {0}.".format(zip_path))
        log.debug(error)
        return False
    else:
        log.debug("Setting comment to ZIP file: {0}.".format(zip_path))
        return True


def get_zip_comment(zip_path):
    """Get a ZIP comment."""
    try:
        log.debug("Getting comment from ZIP file: {0}.".format(zip_path))
        with ZipFile(str(zip_path).strip(), 'r') as myzip:
            return str(myzip.comment.decode("utf-8")).strip()
    except Exception as error:
        log.warning("Failed to get comment from ZIP: {0}.".format(zip_path))
        log.debug(error)
        return ""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Get a random pastel color as string, useful for HTML/CSS styling."""


# This 2 groups have been tested on HTML/CSS with one each other,
# they look pretty good on all combinations, we are not Designers,
# but this is useful for quick templating and boilerplates styling.


from random import choice


def get_random_pastelight_color():
    """Get a random pastel light color as string, useful for CSS styling."""
    return choice((
        'aliceblue', 'antiquewhite', 'aqua', 'aquamarine', 'azure', 'beige',
        'cornsilk', 'floralwhite', 'ghostwhite', 'grey',
        'honeydew', 'ivory', 'lavender',
        'lavenderblush', 'lemonchiffon', 'lightcyan',
        'lightgoldenrodyellow', 'lightgrey', 'lightpink', 'lightskyblue',
        'lightyellow', 'linen', 'mint', 'mintcream', 'oldlace', 'papayawhip',
        'peachpuff', 'seashell', 'skyblue', 'snow', 'thistle', 'white'))


def get_random_pasteldark_color():
    """Get a random dark color as string, useful for CSS styling."""
    return choice((
        'brown', 'chocolate', 'crimson', 'darkblue', 'darkgoldenrod',
        'darkgray', 'darkgreen', 'darkolivegreen', 'darkorange', 'darkred',
        'darkslateblue', 'darkslategray', 'dimgray', 'dodgerblue',
        'firebrick', 'forestgreen', 'indigo', 'maroon', 'mediumblue',
        'midnightblue', 'navy', 'olive', 'olivedrab', 'royalblue',
        'saddlebrown', 'seagreen', 'sienna', 'slategray', 'teal'))


def get_random_pastel_color():
    """Get a random dark or light color as string, useful for CSS styling."""
    return choice((get_random_pastelight_color(),
                   get_random_pasteldark_color()))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Take and Return a string, but Sanitize all weird characters."""


import re


def get_sanitized_string(stringy, repla=""):
    """Take str sanitize non-printable weird characters,return clean str."""
    return re.sub(r"[^\x00-\x7F]+", repla, stringy, flags=re.IGNORECASE)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Determine OS and set copy() and paste() functions accordingly."""


import logging as log
import os
import subprocess
import sys

from shutil import which
from typing import NamedTuple


# Dot notation and static typed and built-in version of collections.namedtuple.
Clipboard = NamedTuple("Clipboard", fields=(("copy", callable),
                                            ("paste", callable)))


def __osx_clipboard():
    def copy_osx(text):
        subprocess.Popen(["pbcopy"], stdin=subprocess.PIPE,
                         close_fds=True).communicate(text.encode("utf-8"))

    def paste_osx():
        os.environ["LANG"] = "en_US.utf-8"
        return str(subprocess.Popen(
            ["pbpaste"], stdout=subprocess.PIPE, close_fds=True
            ).communicate()[0].decode("utf-8"))

    return copy_osx, paste_osx


def __xclip_clipboard():
    def copy_xclip(text):
        subprocess.Popen(["xclip", "-selection", "clipboard"],
                         stdin=subprocess.PIPE, close_fds=True).communicate(
                         text.encode('utf-8'))
        if which("xsel"):
            subprocess.Popen(["xclip", "-selection", "primary"],
                             stdin=subprocess.PIPE, close_fds=True
                             ).communicate(text.encode('utf-8'))

    def paste_xclip():
        return subprocess.Popen(["xclip", "-selection",
                                 "primary" if which("xsel") else "clipboard",
                                 "-o"], stdout=subprocess.PIPE, close_fds=True
                                ).communicate()[0].decode("utf-8")

    return copy_xclip, paste_xclip


def __win32_clibboard():
    import win32clipboard
    import win32con

    def copy_win32(text):
        win32clipboard.OpenClipboard()
        win32clipboard.EmptyClipboard()
        win32clipboard.SetClipboardData(win32con.CF_UNICODETEXT, text)
        win32clipboard.CloseClipboard()

    def paste_win32():
        win32clipboard.OpenClipboard()
        text = win32clipboard.GetClipboardData(win32con.CF_UNICODETEXT)
        win32clipboard.CloseClipboard()
        return text

    return copy_win32, paste_win32


def __determine_clipboard():
    """Determine OS and set copy() and paste() functions accordingly."""
    if sys.platform.startswith("darwin"):
        return __osx_clipboard()
    elif sys.platform.startswith("win"):
        try:  # Determine which command/module is installed, if any.
            import win32clipboard  # lint:ok noqa
            assert win32clipboard
        except ImportError:
            log.error("Install Win32 API Python packages for Windows.")
            return None, None  # install Win32.
        else:
            return __win32_clibboard()
    elif sys.platform.startswith("linux") and which("xclip"):
        return __xclip_clipboard()
    else:
        log.error("Install XClip and XSel Linux Packages at least.")
        return None, None  # install Qt or GTK or Tk or XClip.


def get_clipboard():
    """Crossplatform crossdesktop Clipboard."""
    log.debug("Querying Copy/Paste Clipboard functionality.")
    global clipboard_copy, clipboard_paste
    clipboard_copy, clipboard_paste = None, None
    clipboard_copy, clipboard_paste = __determine_clipboard()
    return Clipboard(clipboard_copy, clipboard_paste)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Decorator for Python3 annotations to type-check inputs and outputs."""


import functools
import logging as log


def typecheck(f):
    """Decorator for Python3 annotations to type-check inputs and outputs."""
    def _check_annotations(tipe):
        _type, is_ok = None, isinstance(tipe, (type, tuple, type(None)))
        if is_ok:  # Annotations can be Type or Tuple or None
            _type = tipe if isinstance(tipe, tuple) else tuple((tipe, ))
            if None in _type:  # if None on tuple replace with type(None)
                _type = tuple([_ if _ is not None else type(_) for _ in _type])
        return _type, is_ok

    @functools.wraps(f)  # wrap a function or method to Type Check it.
    def decorated(*args, **kwargs):
        msg = "Type check error: {0} must be {1} but is {2} on function {3}()."
        notations, f_name = tuple(f.__annotations__.keys()), f.__code__.co_name
        for i, name in enumerate(f.__code__.co_varnames):
            if name not in notations:
                continue  # this arg name has no annotation then skip it.
            _type, is_ok = _check_annotations(f.__annotations__.get(name))
            if is_ok:  # Force to tuple
                if i < len(args) and not isinstance(args[i], _type):
                    log.critical(msg.format(repr(args[i])[:50], _type,
                                            type(args[i]), f_name))
                elif name in kwargs and not isinstance(kwargs[name], _type):
                    log.critical(msg.format(repr(kwargs[name])[:50], _type,
                                            type(kwargs[name]), f_name))
        out = f(*args, **kwargs)
        _type, is_ok = _check_annotations(f.__annotations__.get("return"))
        if is_ok and not isinstance(out, _type) and "return" in notations:
            log.critical(msg.format(repr(out)[:50], _type, type(out), f_name))
        return out    # The output result of function or method.
    return decorated  # The decorated function or method.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Notification message with information, based on D-Bus, with Fallbacks."""


import logging as log

from shutil import which
from subprocess import call

try:
    import dbus
except ImportError:
    dbus = None

try:
    import pynotify
except ImportError:
    pynotify = None


def make_notification(title, message="", name="", icon="", timeout=3000):
    """Notification message with information,based on D-Bus,with Fallbacks."""
    if dbus:  # Theorically the standard universal way.
        log.debug("Sending Notification message via D-Bus API.")
        return dbus.Interface(dbus.SessionBus().get_object(
            "org.freedesktop.Notifications", "/org/freedesktop/Notifications"),
            "org.freedesktop.Notifications").Notify(
                name, 0, icon, title, message, [], [], timeout)
    elif pynotify:  # The non-standard non-universal way.
        log.debug("Sending Notification message via PyNotify API.")
        pynotify.init(name.lower() if name else title.lower())
        return pynotify.Notification(title, message).show()
    elif which("notify-send"):   # The non-standard non-universal sucky ways.
        log.debug("Sending Notification message via notify-send command.")
        comand = (which("notify-send"), "--app-name=" + name,
                  "--expire-time=" + str(timeout), title, message)
        return not bool(call(comand, timeout=timeout // 1000 + 1))
    elif which("kdialog"):
        log.debug("Sending Notification message via KDialog command.")
        comand = (which("kdialog"), "--name=" + name, "--title=" + title,
                  "--icon=" + icon, "--caption=" + name, "--passivepopup",
                  title + message, str(timeout // 1000))
        return not bool(call(comand, timeout=timeout // 1000 + 1))
    elif which("zenity"):
        log.debug("Sending Notification message via Zenity command.")
        comand = (which("zenity"), "--name=" + name, "--title=" + title,
                  "--notification", "--timeout=" + str(timeout // 1000),
                  "--text=" + title + message)
        return not bool(call(comand, timeout=timeout // 1000 + 1))
    elif which("xmessage"):  # This one is Ugly, but I hope you never get here.
        log.warning("Sending Notification message via XMessage command.")
        comand = (which("xmessage"), "-center", title + message)
        return not bool(call(comand, timeout=timeout // 1000 + 1))
    else:  # Windows and Mac dont have API for that, complain to them.
        log.warning("Sending Notifications not supported by this OS.")
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Set process name and cpu priority."""


import logging as log

from ctypes import byref, cdll, create_string_buffer


def set_process_name(name):
    """Set process name and cpu priority."""
    name = str(name).lower().strip()
    try:
        libc = cdll.LoadLibrary("libc.so.6")  # set process name
        buff = create_string_buffer(len(name) + 1)
        buff.value = bytes(name.encode("utf-8"))
        libc.prctl(15, byref(buff), 0, 0, 0)
    except Exception as error:
        log.warning(error)
        return False  # this may fail on windows and its normal, so be silent.
    else:
        log.debug("Process Name set to: {0}.".format(name))
        return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Set or Reset CLI Window Titlebar Title."""


import os
import sys
from shutil import which


def set_terminal_title(titlez=""):
    """Set or Reset CLI Window Titlebar Title."""
    if titlez and isinstance(titlez, str) and len(titlez.strip()):
        if sys.platform.startswith('win'):  # Windows:
            os.system(which("title") + " {0}".format(titlez.strip()))
        else:  # Linux, Os X and otherwise
            print(r"\x1B]0; {0} \x07".format(titlez.strip()))
        return titlez
    else:
        if sys.platform.startswith('win'):
            os.system(which("title"))
        else:
            print(r"\x1B]0;\x07")
        return ""  # Title should be "" so we return ""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#
# To generate DEB package from Python Package:
# sudo pip3 install stdeb
# python3 setup.py --verbose --command-packages=stdeb.command bdist_deb
#
#
# To generate RPM package from Python Package:
# sudo apt-get install rpm
# python3 setup.py bdist_rpm --verbose --fix-python --binary-only
#
#
# To generate EXE MS Windows from Python Package (from MS Windows only):
# python3 setup.py bdist_wininst --verbose
#
#
# To generate PKGBUILD ArchLinux from Python Package (from PyPI only):
# sudo pip3 install git+https://github.com/bluepeppers/pip2arch.git
# pip2arch.py PackageNameHere
#
#
# To Upload to PyPI by executing:
# sudo pip install --upgrade pip setuptools wheel virtualenv
# python3 setup.py register
# python3 setup.py bdist_egg bdist_wheel --universal sdist --formats=bztar,gztar,zip upload --sign


"""Setup.py for Python, as Generic as possible."""


import logging as log
import os
import re
import sys

from setuptools import setup


##############################################################################
# EDIT HERE


MODULE_PATH = os.path.join(os.path.dirname(__file__),
                           "anglerfish", "__init__.py")
DESCRIPTION = """Anglerfish is a simple multipurpose K.I.S.S. Python 3 Helper
Library to help Developers create Apps or Scripts faster, better and easily."""


##############################################################################
# Dont touch below


try:
    with open(str(MODULE_PATH), "r", encoding="utf-8-sig") as source_code_file:
        SOURCE = source_code_file.read()
except:
    with open(str(MODULE_PATH),  "r") as source_code_file:
        SOURCE = source_code_file.read()


def find_this(search, source=SOURCE):
    """Take a string and a filename path string and return the found value."""
    print("Searching for {what}.".format(what=search))
    if not search or not source:
        print("Not found on source: {what}.".format(what=search))
        return ""
    return str(re.compile(r".*__{what}__ = '(.*?)'".format(
        what=search), re.S).match(source).group(1)).strip().replace("'", "")


print("Starting build of setuptools.setup().")


##############################################################################
# EDIT HERE


setup(

    name="anglerfish",
    version=find_this("version"),

    description="Helper Library for Python3 Apps.",
    long_description=DESCRIPTION,

    url=find_this("url"),
    license=find_this("license"),

    author=find_this("author"),
    author_email=find_this("email"),
    maintainer=find_this("author"),
    maintainer_email=find_this("email"),

    include_package_data=True,
    zip_safe=True,

    tests_require=['pytest', 'isort', 'pylama'],

    packages=["anglerfish"],

    keywords=['helper', 'boilerplate', 'utils', 'minimalism', 'utility'],

    classifiers=[

        'Development Status :: 5 - Production/Stable',

        'Environment :: Console',

        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Other Audience',

        'Natural Language :: English',

        'License :: OSI Approved :: GNU General Public License (GPL)',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        'Operating System :: OS Independent',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',

        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',

        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',

        'Topic :: Software Development',

    ],
)


print("Finished build of setuptools.setup().")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Convert bytes to kilobytes, megabytes, gigabytes, etc."""


import os

from collections import OrderedDict
from json import dumps


def walk2dict(folder, links=False, showhidden=False,
              strip=False, jsony=False, ordereddict=False):
    """Return Nested Dictionary represents folder/file structure of folder."""
    ret = []
    for path, dirs, files in os.walk(folder, followlinks=links):
        if not showhidden:
            dirs = [_ for _ in dirs if not _.startswith(".")]
            files = [_ for _ in files if not _.startswith(".")]
        a = {}
        if strip:
            p = path.strip(folder + os.sep)
        else:
            p = path
        if len(p.split(os.sep)) == 1:
            parent = ''
        if len(p.split(os.sep)) > 1:
            parent = os.sep.join(p.split(os.sep)[:-1])
        if path == folder:
            parent = 'root'
        a['path'] = p
        a['fullpath'] = os.path.abspath(path)
        a['parent'] = parent
        a['dirs'] = dirs
        a['files'] = []
        for fyle in files:
            try:  # sometimes os.stat(ff) just fails,breaking all the loop.
                f = {}
                ff = path + os.sep + fyle
                (mode, ino, dev, nlink, uid, gid, size,
                 atime, mtime, ctime) = os.stat(ff)
                f['name'] = fyle
                f['mode'] = mode
                f['ino'] = ino
                f['dev'] = dev
                f['nlink'] = nlink
                f['uid'] = uid
                f['gid'] = gid
                f['size'] = size
                f['atime'] = atime
                f['mtime'] = mtime
                f['ctime'] = ctime
                a['files'].append(f)
            except:
                pass
        ret.append(a)
    dict_of_files = ret[0]
    if ordereddict:  # dictionary sorted by key.
        dict_of_files = OrderedDict(sorted(ret[0].items(), key=lambda t: t[0]))
    if jsony:  # json
        dict_of_files = dumps(dict(dict_of_files), sort_keys=True, indent=4)
    return dict_of_files



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Cross-platform Sound Playing with StdLib only,No Sound file required."""


import logging as log
import os
import sys
from shutil import which
from subprocess import call
from tempfile import gettempdir


def beep(waveform=(79, 45, 32, 50, 99, 113, 126, 127)):
    """Cross-platform Sound Playing with StdLib only,No Sound file required."""
    log.debug("Generating and Playing Sound...")
    wavefile = os.path.join(gettempdir(), "beep.wav")
    if not os.path.isfile(wavefile) or not os.access(wavefile, os.R_OK):
        with open(wavefile, "w+") as wave_file:
            for sample in range(0, 1000, 1):
                for wav in range(0, 8, 1):
                    wave_file.write(chr(waveform[wav]))
    if sys.platform.startswith("linux"):
        repro = which("aplay")
        return not bool(call("{repro} '{fyle}'".format(
            fyle=wavefile, repro=repro), shell=True))
    elif sys.platform.startswith("darwin"):
        repro = which("afplay")
        return not bool(call("{repro} '{fyle}'".format(
            fyle=wavefile, repro=repro), shell=True))
    elif sys.platform.startswith("win"):  # FIXME: This is Ugly.
        log.debug("Playing Sound Natively not supported by this OS.")
        return False  # this SHOULD work on all windows,but it doesnt :(
        # return call("start /low /min '{fyle}'".format(fyle=wavefile),shell=1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Check working folder,from argument,for everything that can go wrong."""


import logging as log
import os

from shutil import disk_usage


def check_folder(folder_to_check=os.path.expanduser("~"), check_space=1):
    """Check working folder,from argument,for everything that can go wrong."""
    folder_to_check = os.path.abspath(folder_to_check)  # More Safe on WinOS
    log.debug("Checking the Working Folder: '{0}'".format(folder_to_check))
    if not isinstance(folder_to_check, str):  # What if folder is not a string.
        log.critical("Folder {0} is not String type!.".format(folder_to_check))
        return False
    elif not os.path.isdir(folder_to_check):  # What if folder is not a folder.
        log.critical("Folder {0} does not exist!.".format(folder_to_check))
        return False
    elif not os.access(folder_to_check, os.R_OK):  # What if not Readable.
        log.critical("Folder {0} not Readable !.".format(folder_to_check))
        return False
    elif not os.access(folder_to_check, os.W_OK):  # What if not Writable.
        log.critical("Folder {0} Not Writable !.".format(folder_to_check))
        return False
    elif disk_usage and os.path.exists(folder_to_check) and bool(check_space):
        hdd = int(disk_usage(folder_to_check).free / 1024 / 1024 / 1024)
        if hdd >= int(check_space):  # >= check_space Gb.
            log.info("Folder Total Free Space: ~{0} GigaBytes.".format(hdd))
            return True
        else:  # < check_space Gb.
            log.critical("Total Free Space: {} GigaBytes.".format(check_space))
            return False
    return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Stealth Strings, hidden and dangerous."""


import base64
import binascii
import codecs
import zlib


def stealth2string(stringy: str, rot13: bool=False) -> str:
    """Stealth to string,stealth is a hidden string,both str type and ttf-8."""

    def __i2b(integ):  # int to bytes, do not touch.
        """Helper for string_to_stealth and stealth_to_string, dont touch!."""
        __num = len("%x" % integ)
        return binascii.unhexlify(str("%x" % integ).zfill(__num + (__num & 1)))

    _n = int(str(stringy).replace(u"\u200B", "0").replace(u"\uFEFF", "1"), 2)
    stringy = zlib.decompress(base64.b64decode(__i2b(_n))).decode('utf-8')
    stringy = codecs.decode(stringy, "rot-13") if rot13 and codecs else stringy
    return str(stringy).strip()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Get a temp Sub-folder for this App only, cross-platform, return path."""


import logging as log
import os

from tempfile import gettempdir


def get_temp_folder(appname):
    """Get a temp Sub-folder for this App only, cross-platform, return path."""
    if appname and len(appname) and isinstance(appname, str):
        temp_path = os.path.join(gettempdir(), appname.strip().lower())
    log.debug("Temp folder for {0} is: {1}.".format(appname, temp_path))
    if not os.path.isdir(temp_path):
        log.debug("Creating new Temp folder: {0}.".format(temp_path))
        os.makedirs(temp_path)
    return temp_path



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


"""Watch a file path for changes run callback if modified. A WatchDog."""


import os
import time


def watch(file_path, callback=None, interval=60, backoff=1,
          timeout=None, repetitions=-1, silent=False, logger=None):
    """Watch a file path for changes run callback if modified. A WatchDog."""
    if not silent:
        msg = "Watching for any changes on file path: {0}.".format(file_path)
        logger.debug(msg) if logger else print(msg)
    previous, file_path = int(os.stat(file_path).st_mtime), str(file_path)
    _interval, msg = interval, "Modification detected on {0}".format(file_path)
    end_time = time.time() + int(timeout) if timeout else None
    repetitions = int(repetitions) if repetitions > 0 else True
    while repetitions:
        if end_time and time.time() > end_time:
            raise  # Timeout!.
        actual = int(os.stat(file_path).st_mtime)
        if previous == actual:
            time.sleep(int(abs(interval)))
            interval *= backoff if backoff and backoff in range(1, 9) else 1
        else:
            previous, interval = actual, _interval
            if not silent:
                logger.debug(msg) if logger else print(msg)
            return callback(file_path) if callback else file_path
